/*   This is part of um-ViewOS
 *   The user-mode implementation of OSVIEW -- A Process with a View
 *
 *   fuse/umfuse module for iso9660 filesystem support
 *   
 *   Copyright 2005 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *   $Id: fuseiso9660.c,v 1.1.1.1 2005/05/19 15:36:30 garden Exp $
 *
 */


#if 0
#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif
#endif

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <cdio/iso9660.h>
#include <cdio/logging.h>

#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif

int count=100;
static int f_iso9660_getattr(const char *path, struct stat *stbuf)
{
	struct fuse_context *mycontext=fuse_get_context();
	iso9660_t *isofs=(iso9660_t *) mycontext->private_data;

	//printf("f_iso9660_getattr %s\n",path);
	iso9660_stat_t *isostat=
		//iso9660_ifs_stat_translate(isofs,path);
		iso9660_ifs_stat(isofs,path);
	//printf("f_iso9660_getattr %s done\n",path);
	if (isostat ==NULL)
		return -ENOENT;
	else {
		memset (stbuf,0,sizeof(struct stat));
		stbuf->st_dev=0;
		stbuf->st_ino=count++;
		switch (isostat->type) {
			case _STAT_FILE: 
				stbuf->st_mode=__S_IFREG|0444;break;
			case _STAT_DIR: 
				stbuf->st_mode=__S_IFDIR|0555;break;
		}
		stbuf->st_nlink=1;
		stbuf->st_uid=0;
		stbuf->st_gid=0;
		stbuf->st_rdev=0;
		stbuf->st_size=isostat->size;
		stbuf->st_blksize=ISO_BLOCKSIZE;
		stbuf->st_blocks=isostat->secsize;
		stbuf->st_atime= stbuf->st_mtime= stbuf->st_ctime=mktime(&isostat->tm);
		char filename[4096];
		//iso9660_name_translate(isostat->filename, filename);
		//iso9660_name_translate_ext(isostat->filename, filename, iso9660_ifs_get_joliet_level(isofs));
		strcpy(filename,isostat->filename);
		//printf("f_iso9660_getattr OKAY %s\n",filename);
		return 0;
	}
}

static int f_iso9660_readlink(const char *path, char *buf, size_t size)
{
    return -EINVAL;
}

static int f_iso9660_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
	struct fuse_context *mycontext=fuse_get_context();
	iso9660_t *isofs=(iso9660_t *) mycontext->private_data;

	//printf("f_iso9660_getdir %s\n",path);
	iso9660_stat_t *isostat=
		//iso9660_ifs_stat_translate(isofs,path);
		iso9660_ifs_stat(isofs,path);
	//printf("f_iso9660_getdir %s stat done\n",path);
	if (isostat ==NULL)
		return -ENOENT;
	else {
		CdioList *entlist;
		CdioListNode *entnode;

	//printf("f_iso9660_readdir %s\n",path);
		entlist = iso9660_ifs_readdir (isofs, path);
	//printf("f_iso9660_readdir %s done\n",path);

		if (entlist != NULL) {

			_CDIO_LIST_FOREACH (entnode, entlist)
			{
				char filename[4096];
				int type;
				iso9660_stat_t *p_statbuf = _cdio_list_node_data (entnode);
	//printf("f_iso9660_translate %s \n",path);
				//iso9660_name_translate(p_statbuf->filename, filename);
				//iso9660_name_translate_ext(p_statbuf->filename, filename, iso9660_ifs_get_joliet_level(isofs));
				strcpy(filename,p_statbuf->filename);
	//printf("f_iso9660_translate %s done\n",path);
				//printf ("/%s\n", filename);
				switch (p_statbuf->type) {
					case _STAT_FILE: type=DT_REG;break;
					case _STAT_DIR:  type=DT_DIR;break;
				}
				//printf("filler %s %d\n",filename,type);
				filler(h, filename, type, 2);
			}

			_cdio_list_free (entlist, true);
		}

	}
	return 0;
}

#if 0
static int f_iso9660_mknod(const char *path, mode_t mode, dev_t rdev)
{
    int res;

    res = mknod(path, mode, rdev);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_mkdir(const char *path, mode_t mode)
{
    int res;

    res = mkdir(path, mode);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_unlink(const char *path)
{
    int res;

    res = unlink(path);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_rmdir(const char *path)
{
    int res;

    res = rmdir(path);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_symlink(const char *from, const char *to)
{
    int res;

    res = symlink(from, to);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_rename(const char *from, const char *to)
{
    int res;

    res = rename(from, to);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_link(const char *from, const char *to)
{
    int res;

    res = link(from, to);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_chmod(const char *path, mode_t mode)
{
    int res;

    res = chmod(path, mode);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_chown(const char *path, uid_t uid, gid_t gid)
{
    int res;

    res = lchown(path, uid, gid);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_truncate(const char *path, off_t size)
{
    int res;

    res = truncate(path, size);
    if(res == -1)
        return -errno;

    return 0;
}

static int f_iso9660_utime(const char *path, struct utimbuf *buf)
{
    int res;

    res = utime(path, buf);
    if(res == -1)
        return -errno;

    return 0;
}
#endif


static int f_iso9660_open(const char *path, struct fuse_file_info *fi)
{
    return 0;
}

static int f_iso9660_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
	//printf("f_iso9660_read %s\n",path);
	struct fuse_context *mycontext=fuse_get_context();
	iso9660_t *isofs=(iso9660_t *) mycontext->private_data;

	iso9660_stat_t *isostat=
		//iso9660_ifs_stat_translate(isofs,path);
		iso9660_ifs_stat(isofs,path);
	//printf("f_iso9660_read %s stat done\n",path);
	if (isostat ==NULL)
		return -ENOENT;
	else {
		off_t firstblock=offset/ISO_BLOCKSIZE;
		off_t lastblock;
		int readrv=1;
		off_t bytesread=0;
		if (offset+size > isostat->size)
			size= isostat->size - offset;
		lastblock=(offset+size)/ISO_BLOCKSIZE;
		char *cdbuf[ISO_BLOCKSIZE];
		memset (cdbuf, 0, ISO_BLOCKSIZE);
		offset -= firstblock*ISO_BLOCKSIZE;
		while (size > 0 && readrv > 0) {
	//printf("f_iso9660_read %s seek/read \n",path);
			readrv=iso9660_iso_seek_read(isofs, cdbuf, 
					isostat->lsn+firstblock,1);
	//printf("f_iso9660_read %s seek/read done\n",path);
			if (readrv>0) {
				off_t len=(ISO_BLOCKSIZE-offset < size)?ISO_BLOCKSIZE-offset:size;
				memcpy(buf,cdbuf+offset,len);
				offset=0;
				size-=len;
				buf+=len;
				bytesread+=len;
				firstblock++;
			}
		} 
		return bytesread;
	}
}

#if 0 
static int f_iso9660_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    int fd;
    int res;

    (void) fi;
    fd = open(path, O_WRONLY);
    if(fd == -1)
        return -errno;

    res = pwrite(fd, buf, size, offset);
    if(res == -1)
        res = -errno;

    close(fd);
    return res;
}

static int f_iso9660_statfs(const char *path, struct statfs *stbuf)
{
    int res;

    res = statfs(path, stbuf);
    if(res == -1)
        return -errno;

    return 0;
}
#endif

static int f_iso9660_release(const char *path, struct fuse_file_info *fi)
{
    return 0;
}

#if 0
static int f_iso9660_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    /* Just a stub.  This method is optional and can safely be left
       unimplemented */

    (void) path;
    (void) isdatasync;
    (void) fi;
    return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int f_iso9660_setxattr(const char *path, const char *name, const char *value,
                        size_t size, int flags)
{
    int res = lsetxattr(path, name, value, size, flags);
    if(res == -1)
        return -errno;
    return 0;
}

static int f_iso9660_getxattr(const char *path, const char *name, char *value,
                    size_t size)
{
    int res = lgetxattr(path, name, value, size);
    if(res == -1)
        return -errno;
    return res;
}

static int f_iso9660_listxattr(const char *path, char *list, size_t size)
{
    int res = llistxattr(path, list, size);
    if(res == -1)
        return -errno;
    return res;
}

static int f_iso9660_removexattr(const char *path, const char *name)
{
    int res = lremovexattr(path, name);
    if(res == -1)
        return -errno;
    return 0;
}
#endif /* HAVE_SETXATTR */
#endif

static struct fuse_operations iso9660_oper = {
    .getattr	= f_iso9660_getattr,
    .readlink	= f_iso9660_readlink,
    .getdir	= f_iso9660_getdir,
    .open	= f_iso9660_open,
    .read	= f_iso9660_read,
    .release	= f_iso9660_release,
/*
ok    .getattr	= f_iso9660_getattr,
ok    .readlink	= f_iso9660_readlink,
ok    .getdir	= f_iso9660_getdir,
    .mknod	= f_iso9660_mknod,
    .mkdir	= f_iso9660_mkdir,
    .symlink	= f_iso9660_symlink,
    .unlink	= f_iso9660_unlink,
    .rmdir	= f_iso9660_rmdir,
    .rename	= f_iso9660_rename,
    .link	= f_iso9660_link,
    .chmod	= f_iso9660_chmod,
    .chown	= f_iso9660_chown,
    .truncate	= f_iso9660_truncate,
    .utime	= f_iso9660_utime,
ok    .open	= f_iso9660_open,
ok    .read	= f_iso9660_read,
    .write	= f_iso9660_write,
    .statfs	= f_iso9660_statfs,
ok    .release	= f_iso9660_release,
    .fsync	= f_iso9660_fsync,
    */
#if 0
#ifdef HAVE_SETXATTR
    .setxattr	= f_iso9660_setxattr,
    .getxattr	= f_iso9660_getxattr,
    .listxattr	= f_iso9660_listxattr,
    .removexattr= f_iso9660_removexattr,
#endif
#endif
};


struct fuse *fuse;
int fuse_fd;

int main(int argc, char *argv[])
{
	int err;
	struct fuse_context *mycontext;
	iso9660_t *isofs;

	if (argc < 3)
		return -1;
	//if (strcmp(argv[1],"-o")==0)
	//	argv+=2;
	//isofs=iso9660_open(argv[1]);
	isofs=iso9660_open_ext(argv[argc-2],ISO_EXTENSION_ALL);
	iso9660_ifs_read_superblock(isofs,ISO_EXTENSION_ALL);
	cdio_loglevel_default=CDIO_LOG_ERROR;
	if (isofs!=NULL) {
		fuse_fd = fuse_mount(argv[argc-1], "ro");
		fuse = fuse_new(fuse_fd, NULL, &iso9660_oper, sizeof(iso9660_oper));
		//printf("MOUNT OKAY!\n");
		mycontext=fuse_get_context();
		mycontext->private_data=isofs;

		fuse_loop(fuse);

		iso9660_close(isofs);
		return 0;
	}
	else
		return -1;
}
